# Copyright (C) 2007 Richard Musiol
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

require 'inline'

class String
  def to_class
    names = self.split('::')
    klass = Object
    names.each do |name|
      sym = name.to_sym
      return nil if !klass.const_defined?(sym)
      klass = klass.const_get(sym)
    end
    klass
  end
end

class Module
  def parent_module
    self.name.split('::')[0..-2].join('::').to_class
  end
  
  def lazy_attr_reader(name, &block)
    attr_name = name.to_s.gsub('?', '')
    define_method("lazy_attr_initialization_#{attr_name}".to_sym, block)
    class_eval <<-CODE_END
      def #{name.to_s}
        @#{attr_name} = lazy_attr_initialization_#{attr_name} if @#{attr_name}.nil?
        @#{attr_name}
      end
    CODE_END
  end
end

class Array
  def all_indices
   (0...self.size).to_a
  end
  
  def assigned_indices
    all_indices.reject { |i| self[i].nil? }
  end
  
  def prefix(str)
    self.map { |item| "#{str}#{item}" }
  end
  
  def prefix!(str)
    self.map! { |item| "#{str}#{item}" }
  end
  
  def postfix(str)
    self.map { |item| "#{item}#{str}" }
  end
  
  def postfix!(str)
    self.map! { |item| "#{item}#{str}" }
  end
  
  def map_with_indices(target = [])
    self.each_index do |index|
      target[index] = yield(self[index], index)
    end
    target
  end
  
  def map_with_indices!
    self.map_with_indices(self)
  end
  
  def map_method_results(method_name)
    self.map { |item| item.send(method_name) }
  end
  
  def map_method_results!(method_name)
    self.map! { |item| item.send(method_name) }
  end
  
  alias_method :throw_method_missing, :method_missing
  def method_missing(name, *args, &block)
    if name.to_s[-1..-1] == 's'
      method_name = name.to_s[0..-2].to_sym
      if size == 0 || self[0].respond_to?(method_name)
        return map_method_results(method_name)
      end
    end
    throw_method_missing(name, args, &block)
  end
  
  def reject_nils
    self.reject { |item| item.nil? }
  end
  
  def reject_nils!
    self.reject! { |item| item.nil? }
  end
end

module Inline
  class C
    #    TRANSFER_TYPE = $TESTING ? 'unsigned long' : 'VALUE'
    TRANSFER_TYPE = 'long'
    
    @@internal_c_functions = []
    @@internal_c_functions_module = nil
    
    def self.internal_c_functions
      @@internal_c_functions
    end
    
    def self.internal_c_functions_module
      @@internal_c_functions_module
    end
    
    def self.type_map
      @@type_map
    end
    
    alias_method :initialize_orig, :initialize
    def initialize(mod)
      initialize_orig(mod)
      if @@internal_c_functions_module.nil?
        @@internal_c_functions.each { |function| self.c_singleton(function) }
        @@internal_c_functions_module = mod
      end
      @eval_on_load = ''
    end
    
    alias_method :load_orig, :load
    def load
      load_orig
      @mod.class_eval(@eval_on_load)
    end
    
    def link_function(c_return_type, function_name, parameter_types)
      delegate_function_name = "C_delegate_#{function_name}"
      return_type = Type.new(c_return_type, link_type(c_return_type))
      parameters = parameter_types.map_with_indices { |type, index| Parameter.new(type, link_type(type), index)  }
      
      #build c delegate
      c_code = <<-CODE_END
        #{return_type.c_transfer_type} #{delegate_function_name}(#{parameters.c_method_parameters.join(', ')}) {
          #{return_type.void? ? '' : "return (#{return_type.c_transfer_type})"} #{function_name}(#{parameters.c_casted_call_args.join(', ')});
        }
        CODE_END
      #      puts c_code
      self.c_singleton c_code
      
      #build ruby delegate
      line = __LINE__; ruby_code = <<-CODE_END
        def self.#{function_name}(#{parameters.ruby_method_parameters.join(', ')})
          #{parameters.ruby_ref_assigns.reject_nils.join('; ')}
          return_values = []
          #{return_type.void? ? '' : ('return_values << ' + return_type.ruby_type_conversion)}(#{delegate_function_name}(#{parameters.ruby_transformed_call_args.join(', ')}))
          #{parameters.ruby_converted_return_values.reject_nils.prefix('return_values << ').join('; ')}
          case return_values.size
            when 0: nil
            when 1: return_values[0]
            else    return_values
          end
        end
        def #{function_name}(#{parameters.ruby_method_parameters.join(', ')})
          self.class.#{function_name}(#{parameters.arg_names.join(', ')})
        end
        CODE_END
      #      puts ruby_code
      @mod.class_eval ruby_code, __FILE__, line + 1
    end
    
    def link_fields(struct_name, field_hash)
      field_hash.each do |field_name, c_type|
        delegate_function_name = "#{struct_name.gsub(' ', '_')}_#{field_name}"
        type = Type.new(c_type, link_type(c_type))
        
        #build c delegate
        c_code = <<-CODE_END
          #{type.c_transfer_type} #{delegate_function_name}(#{TRANSFER_TYPE} pointer) {
            #{type.c_transfer_type} result = (#{type.c_transfer_type}) ((#{struct_name}*) pointer)->#{field_name};
            #{type.is_simple_type? ? 'return result;' : "if(result) { return result; } else { return(Qfalse); }"}
          }
          CODE_END
        #        puts c_code
        self.c_singleton c_code
        
        #build ruby delegate
        line = __LINE__; ruby_code = <<-CODE_END
          def #{field_name}
            result = (#{@mod.name}.#{delegate_function_name}(memory_address))
            #{type.is_pointer_type? ? 'return nil if !result' : ''}
            #{type.ruby_type_conversion}(result)
          end
          CODE_END
        #        puts ruby_code
        link_type("#{struct_name}*").class_eval ruby_code, __FILE__, line + 1
      end
    end
    
    def link_type(c_type, reallocate_pointers=false, &block)
      is_simple_type = @@type_map.has_key?(c_type)
      is_pointer = c_type[-1..-1] == '*'
      return nil if is_simple_type
      return Value if !is_pointer
      
      c_inner_type = c_type[0..-2].strip
      class_name = c_inner_type.gsub(' ', '_').capitalize
      klass = "#{@mod.name}::#{class_name}".to_class
      if !klass
        inner_type = Type.new(c_inner_type, nil)
        is_simple_type = @@type_map.has_key?(c_inner_type)
        
        klass = Class.new()
        @mod.const_set(class_name.to_sym, klass)

        line = __LINE__; ruby_code = <<-CODE_END
          include Inline::C::Pointer
         
          def dereference
            #{link_type('void *').name}.new(Inline::C.internal_c_functions_module.C_Pointer_dereference_to_pointer(memory_address))
          end
          
          alias_method :throw_method_missing, :method_missing
          def method_missing(name, *args, &block)
            if name.to_s[0..2] == 'to_'
              type = name.to_s[3..-1]
              klass = ("#{@mod.name}::" + type.capitalize).to_class
              return klass.new(memory_address, ruby_managed)
            end
            throw_method_missing(name, *args, &block)
          end
          CODE_END
        klass.class_eval ruby_code, __FILE__, line + 1
        
        if !inner_type.void?
          create_pointer_from_data_function_name = "#{class_name}_create_pointer_from_data"
          create_pointer_from_pointer_function_name = "#{class_name}_create_pointer_from_pointer"
          get_array_entry_function_name = "#{class_name}_get_array_entry"
        
          #create pointer from data
          c_code = <<-CODE_END
            #{TRANSFER_TYPE} #{create_pointer_from_data_function_name}(#{is_simple_type ? c_inner_type : TRANSFER_TYPE} source) {
              #{c_type} pointer = (#{c_type}) malloc(sizeof(#{c_inner_type}));
              memcpy(pointer, &source, sizeof(#{c_inner_type}));
              return (#{TRANSFER_TYPE}) pointer;
            }
            CODE_END
          self.c_singleton c_code
          
          #create pointer from pointer
          c_code = <<-CODE_END
            #{TRANSFER_TYPE} #{create_pointer_from_pointer_function_name}(#{TRANSFER_TYPE} source) {
              #{c_type} pointer = (#{c_type}) malloc(sizeof(#{c_inner_type}));
              *pointer = *((#{c_type}) source);
              return (#{TRANSFER_TYPE}) pointer;
            }
            CODE_END
          self.c_singleton c_code
          
          #array delegate
          c_code = <<-CODE_END
            #{TRANSFER_TYPE} #{get_array_entry_function_name}(#{TRANSFER_TYPE} source, long index) {
              return (#{TRANSFER_TYPE}) &((#{c_type}) source)[index];
            }
            CODE_END
          self.c_singleton c_code
          
          line = __LINE__; ruby_code = <<-CODE_END
            @@reallocate_pointers = #{reallocate_pointers}
            def self.reallocate_pointers?
              @@reallocate_pointers
            end
            
            def self.create(source=0)
              if source.is_a? Inline::C::Pointer
                if !@@reallocate_pointers
                  source
                else
                  self.new(#{@mod.name}.#{create_pointer_from_pointer_function_name}(source.memory_address), true)
                end
              else
                self.new(#{@mod.name}.#{create_pointer_from_data_function_name}(Inline::C::Value.get_data(source)), true)
              end
            end
          
            def [](index)
              self.class.new(#{@mod.name}.#{get_array_entry_function_name}(self.memory_address, index), self.ruby_managed)
            end
          CODE_END
          klass.class_eval ruby_code, __FILE__, line + 1
        end
      end
      
      klass.class_eval(&block) if block
      klass
    end
    
    def link_constants(type, names)
      names.each do |name|
        get_constant_method = "get_constant_#{name}"
        
        self.c_singleton <<-CODE_END
          #{type} #{get_constant_method}(void) {
            return #{name};
          }
          CODE_END
        
        @eval_on_load << <<-CODE_END
          #{@mod.name}.const_set(:#{name}, #{get_constant_method})
          CODE_END
      end
    end
    
    class Type
      def initialize(c_type, ruby_type_class)
        @c_type = c_type
        @ruby_type_class = ruby_type_class
      end
      
      attr_reader :c_type, :ruby_type_class
      
      lazy_attr_reader :is_simple_type? do
        void? || C.type_map.has_key?(c_type)
      end
      
      lazy_attr_reader :is_pointer_type? do
        !is_simple_type? && c_type[-1..-1] == '*'
      end
      
      def void?
        c_type == 'void'
      end
      
      def c_transfer_type
        is_simple_type? ? c_type : TRANSFER_TYPE
      end
      
      def ruby_type_conversion
        is_simple_type? ? '' : "#{ruby_type_class.name}.new"
      end
    end
    
    class Parameter < Type
      alias_method :type_initialize, :initialize
      def initialize(c_type, ruby_type_class, index)
        type_initialize(c_type, ruby_type_class)
        @index = index
      end
      
      attr_reader :index
      
      def arg_name
        "arg#{index}"
      end
      
      def ref_name
        "ref#{index}"
      end
      
      def c_method_parameter
        "#{c_transfer_type} #{arg_name}"
      end
      
      def c_casted_call_arg
        "(#{c_type}) #{arg_name}"
      end
      
      def ruby_method_parameter
        "#{arg_name}=" + (is_pointer_type? ? 'nil' : '0')
      end
      
      def ruby_ref_assign
        return nil if !is_pointer_type?
        "#{ref_name} = #{ruby_type_class.name}.create(#{arg_name})"
      end
      
      def ruby_transformed_call_arg
        is_pointer_type? ? "#{ref_name}.memory_address" : "Inline::C::Value.get_data(#{arg_name})"
      end
      
      def ruby_converted_return_value
        return nil if !is_pointer_type?
       (is_simple_type? ? "#{ref_name}.to_#{c_type[0..-2].strip.gsub(' ', '_')}" : ref_name + (ruby_type_class.reallocate_pointers? ? '' : " if #{arg_name}.nil?"))
      end
    end
    
    class Value
      def initialize(data)
        @data = ValueSafe.new()
        @data.value = data
      end
      
      def data
        @data.value
      end
      
      def self.get_data(source)
        if source.is_a?(self) 
          source.data
        elsif source.nil?
          0
        else
          source
        end
      end
      
      ['char', 'double', 'int', 'long', 'unsigned int', 'unsigned long'].each do |type|
        Inline::C.internal_c_functions << "#{type} C_Value_convert_to_#{type.gsub(' ', '_')}(#{Inline::C::TRANSFER_TYPE} data) { return (#{type}) data; }"
        
        #actual conversion is done by RubyInline
        self.class_eval <<-CODE_END
          def to_#{type.gsub(' ', '_')}
            return Inline::C.internal_c_functions_module.C_Value_convert_to_#{type.gsub(' ', '_')}(data)
          end
          CODE_END
      end
    end
    
    module Pointer
      @@registered_pointers = {}
      
      attr_reader :ruby_managed
      
      def initialize(memory_address, ruby_managed=false)
        @memory_address = ValueSafe.new()
        @memory_address.value = memory_address
        @ruby_managed = ruby_managed
        
        if ruby_managed
          @@registered_pointers[self.object_id] = memory_address
          ObjectSpace.define_finalizer(self, lambda { |id|
            memory_address = @@registered_pointers.delete(id)
            if !@@registered_pointers.has_value?(memory_address)
              Inline::C.internal_c_functions_module.C_Pointer_free_memory(memory_address)
            end
          })
        end
      end
      
      def memory_address
        @memory_address.value
      end
      
      def ==(other)
       (other.is_a? Pointer) && Inline::C.internal_c_functions_module.C_Pointer_equals(self.memory_address, other.memory_address)
      end
      
      Inline::C.internal_c_functions <<  "void C_Pointer_free_memory(#{Inline::C::TRANSFER_TYPE} pointer) { free((void*) pointer); }"
      Inline::C.internal_c_functions <<  "#{Inline::C::TRANSFER_TYPE} C_Pointer_dereference_to_pointer(#{Inline::C::TRANSFER_TYPE} pointer) { return (#{Inline::C::TRANSFER_TYPE}) *((void**) pointer); }"
      Inline::C.internal_c_functions <<  "VALUE C_Pointer_equals(#{Inline::C::TRANSFER_TYPE} pointer1, #{Inline::C::TRANSFER_TYPE} pointer2) { return (pointer1 == pointer2) ? Qtrue : Qfalse; }"
      
      ['char', 'double', 'int', 'long', 'unsigned int', 'unsigned long'].each do |type|
        Inline::C.internal_c_functions << "#{type} C_Pointer_dereference_to_#{type.gsub(' ', '_')}(#{Inline::C::TRANSFER_TYPE} pointer) { return *((#{type}*) pointer); }"
        
        self.class_eval <<-CODE_END
          def to_#{type.gsub(' ', '_')}
            return Inline::C.internal_c_functions_module.C_Pointer_dereference_to_#{type.gsub(' ', '_')}(memory_address)
          end
        CODE_END
      end
    end
  end
  
  class ValueSafe
    def value
      @value
    end
    
    def value=(data)
      @value = data
    end
    
    def to_s
      $TESTING ? @value : '[enable $TESTING]'
    end
    
    def inspect
      to_s
    end
  end
end
